from datetime import datetime, timedelta
import enum
import os
from pathlib import Path
from typing import Optional

from dotenv import find_dotenv, load_dotenv
from rich.console import Console
from rich.table import Table
from tqdm import tqdm
import typer
from loguru import logger

from cloudpathlib import implementation_registry


from runner import main, normalize_results, results_to_csv

# make loguru and tqdm play nicely together
logger.remove()
logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)

# get environement variables
load_dotenv(find_dotenv())

# enumerate cloudpathlib implementations
CloudEnum = enum.Enum("CloudEnum", {k: k for k in implementation_registry.keys()})

# initialize CLI
cli = typer.Typer()


def _configure(
    backend: CloudEnum,
    bucket: Optional[str] = None,
):
    # common configuration for all commands
    logger.info(f"Setting up tests with backend {backend}.")

    if bucket is None:
        logger.info("Bucket not set explicitly, loading from environment variable.")
        bucket = {
            "s3": os.environ.get("LIVE_S3_BUCKET"),
            "gs": os.environ.get("LIVE_AZURE_CONTAINER"),
            "azure": os.environ.get("LIVE_GS_BUCKET"),
        }.get(backend.value)

    logger.info(f"Bucket: {bucket}")

    # get the actual implementation
    backend = implementation_registry.get(backend.value)

    return backend.path_class(f"{backend.path_class.cloud_prefix}{bucket}/performance_tests")


def results_to_rich_table(results):
    table = Table(title=f"Performance suite results: ({datetime.now().isoformat().split('Z')[0]})")

    var_to_title = lambda x: (" ".join(x.split("_"))).title()

    all_fields, row_list = normalize_results(results)

    for field in all_fields:
        col_kwargs = {}
        # get a sample value
        val = row_list[0][field]

        if isinstance(val, (int, float)):
            col_kwargs["justify"] = "right"

        # average performance value is highlighted in green
        if field == "mean":
            col_kwargs["style"] = "green"

        table.add_column(var_to_title(field), **col_kwargs)

    def _format_row(r):
        formatted = []
        for f in all_fields:
            val = r[f]

            if f in ["mean", "min", "max", "std"]:
                val = str(timedelta(seconds=val))
            elif isinstance(val, int):
                val = f"{val:,}"

            if f == "std":
                val = "± " + val

            formatted.append(val)

        return formatted

    for row in row_list:
        table.add_row(*_format_row(row))

    return table


@cli.command(short_help="Runs peformance test suite against a specific backend and bucket.")
def run(
    backend: CloudEnum,
    bucket: Optional[str] = None,
    iterations: int = 10,
    burn_in: int = 2,
    save_csv: Optional[Path] = None,
):
    root = _configure(backend, bucket)

    results = main(root, iterations, burn_in)

    c = Console()
    c.print(results_to_rich_table(results))

    if save_csv is not None:
        results_to_csv(results, save_csv)


if __name__ == "__main__":
    cli()



import csv
from dataclasses import dataclass
from pathlib import Path
from time import perf_counter
from typing import Callable, Dict, List

from statistics import mean, stdev


from tqdm import tqdm
from tqdm.contrib.concurrent import thread_map
from loguru import logger

from cloudpathlib import CloudPath


from perf_file_listing import folder_list


# make loguru and tqdm play nicely together
logger.remove()
logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)


def construct_tree(folder_depth, sub_folders, items_per_folder):
    for limit in range(0, folder_depth + 1):
        if limit == 0:
            prefix = ""
        else:
            prefix = "/".join(f"level_{d:05}" for d in range(1, limit + 1))
            prefix += "/"  # append slash for when this gets combined later

        for f in range(0, sub_folders + 1):
            if f == 0:
                folder_prefix = prefix
            else:
                folder_prefix = f"{prefix}folder_{f:05}/"

            for i in range(1, items_per_folder + 1):
                yield f"{folder_prefix}{i:05}.item"


def setup_test(root, folder_depth, sub_folders, items_per_folder, overwrite=False):
    test_config_str = f"{folder_depth}_{sub_folders}_{items_per_folder}"
    test_folder = CloudPath(root) / test_config_str

    if test_folder.exists():
        if not overwrite:
            logger.info(
                f"Folder '{test_folder}' already exists, setup complete. Pass 'overwrite=True' to delete and rewrite"
            )
            return test_folder
        else:
            logger.info(
                f"Folder '{test_folder}' already exists, and overwrite=True. Removing existing folder."
            )
            test_folder.rmtree()

    logger.info(f"Setting up files for testing in folder: {test_folder}")

    # create folders and files for test in parallel
    thread_map(
        lambda x: (test_folder / x).touch(),
        list(construct_tree(folder_depth, sub_folders, items_per_folder)),
        desc="creating...",
    )

    return test_folder


@dataclass
class PerfRunConfig:
    name: str
    args: List
    kwargs: Dict


def run_single_perf_test(
    func: Callable, iterations: int, burn_in: int, configs: List[PerfRunConfig]
):
    all_results = {}
    for c in configs:
        measurements = []
        for i in tqdm(range(iterations + burn_in), desc=c.name):
            t0 = perf_counter()

            result = func(*c.args, **c.kwargs)

            if i >= burn_in:
                measurements.append(perf_counter() - t0)

        stats = {}
        stats["mean"] = mean(measurements)
        stats["max"] = max(measurements)
        stats["std"] = stdev(measurements)
        stats["iterations"] = iterations

        if isinstance(result, dict):
            stats.update(result)
        else:
            stats["result"] = result

        all_results[c.name] = stats  # add any return information from the test itself

    return all_results


def main(root, iterations, burn_in):
    # required folder setups; all totals over ~5,000 so that we get
    # automatically paginated by AWS S3 API
    shallow = setup_test(root, 0, 0, 5_500)
    normal = setup_test(root, 5, 100, 12)
    deep = setup_test(root, 50, 5, 25)

    test_suite = [
        (
            "List Folders",
            folder_list,
            [
                PerfRunConfig(name="List shallow recursive", args=[shallow, True], kwargs={}),
                PerfRunConfig(name="List shallow non-recursive", args=[shallow, False], kwargs={}),
                PerfRunConfig(name="List normal recursive", args=[normal, True], kwargs={}),
                PerfRunConfig(name="List normal non-recursive", args=[normal, False], kwargs={}),
                PerfRunConfig(name="List deep recursive", args=[deep, True], kwargs={}),
                PerfRunConfig(name="List deep non-recursive", args=[deep, False], kwargs={}),
            ],
        ),
    ]

    logger.info(
        f"Running performance test suite: root={root}, iter={iterations}, burn_in={burn_in}"
    )

    all_results = {}

    for name, func, confs in tqdm(test_suite, desc="Tests"):
        all_results[name] = run_single_perf_test(func, iterations, burn_in, confs)

    return all_results


def normalize_results(results):
    """Convert nested dict of results to a list of dicts with all the same keys."""
    rows_list = []
    observed_fields = set()

    for test_name, test_results in results.items():
        for conf_name, conf_results in test_results.items():
            row_dict = {"test_name": test_name, "config_name": conf_name}
            row_dict.update(conf_results)
            rows_list.append(row_dict)

            observed_fields |= set(row_dict.keys())

    # order fields for ones present in all tests; then any extra fields at the end
    common_fields = ["test_name", "config_name", "iterations", "mean", "std", "max"]

    all_fields = common_fields + list(observed_fields - set(common_fields))

    return all_fields, rows_list


def results_to_csv(results, path):
    """Save the results from `runner.main` to a csv file."""
    all_fields, row_list = normalize_results(results)

    with Path(path).open("w", newline="") as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=all_fields)
        writer.writeheader()

        for row in row_list:
            # normalize so all rows have all possible fields
            for field in all_fields:
                if field not in row:
                    row[field] = None

            writer.writerow(row)



import abc
import mimetypes
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generic, Callable, Iterable, Optional, Tuple, TypeVar, Union

from .cloudpath import CloudImplementation, CloudPath, implementation_registry

BoundedCloudPath = TypeVar("BoundedCloudPath", bound=CloudPath)


def register_client_class(key: str) -> Callable:
    def decorator(cls: type) -> type:
        if not issubclass(cls, Client):
            raise TypeError("Only subclasses of Client can be registered.")
        implementation_registry[key]._client_class = cls
        implementation_registry[key].name = key
        cls._cloud_meta = implementation_registry[key]
        return cls

    return decorator


class Client(abc.ABC, Generic[BoundedCloudPath]):
    _cloud_meta: CloudImplementation
    _default_client = None

    def __init__(
        self,
        local_cache_dir: Optional[Union[str, os.PathLike]] = None,
        content_type_method: Optional[Callable] = mimetypes.guess_type,
    ):
        self._cache_tmp_dir = None
        self._cloud_meta.validate_completeness()
        # setup caching and local versions of file and track if it is a tmp dir
        if local_cache_dir is None:
            self._cache_tmp_dir = TemporaryDirectory()
            local_cache_dir = self._cache_tmp_dir.name

        self._local_cache_dir = Path(local_cache_dir)
        self.content_type_method = content_type_method

    def __del__(self) -> None:
        # make sure temp is cleaned up if we created it
        if self._cache_tmp_dir is not None:
            self._cache_tmp_dir.cleanup()

    @classmethod
    def get_default_client(cls) -> "Client":
        """Get the default client, which the one that is used when instantiating a cloud path
        instance for this cloud without a client specified.
        """
        if cls._default_client is None:
            cls._default_client = cls()
        return cls._default_client

    def set_as_default_client(self) -> None:
        """Set this client instance as the default one used when instantiating cloud path
        instances for this cloud without a client specified."""
        self.__class__._default_client = self

    def CloudPath(self, cloud_path: Union[str, BoundedCloudPath]) -> BoundedCloudPath:
        return self._cloud_meta.path_class(cloud_path=cloud_path, client=self)  # type: ignore

    @abc.abstractmethod
    def _download_file(
        self, cloud_path: BoundedCloudPath, local_path: Union[str, os.PathLike]
    ) -> Path:
        pass

    @abc.abstractmethod
    def _exists(self, cloud_path: BoundedCloudPath) -> bool:
        pass

    @abc.abstractmethod
    def _list_dir(
        self, cloud_path: BoundedCloudPath, recursive: bool
    ) -> Iterable[Tuple[BoundedCloudPath, bool]]:
        """List all the files and folders in a directory.

        Parameters
        ----------
        cloud_path : CloudPath
            The folder to start from.
        recursive : bool
            Whether or not to list recursively.

        Returns
        -------
        contents : Iterable[Tuple]
            Of the form [(CloudPath, is_dir), ...] for every child of the dir.
        """
        pass

    @abc.abstractmethod
    def _move_file(
        self, src: BoundedCloudPath, dst: BoundedCloudPath, remove_src: bool = True
    ) -> BoundedCloudPath:
        pass

    @abc.abstractmethod
    def _remove(self, path: BoundedCloudPath, missing_ok: bool = True) -> None:
        """Remove a file or folder from the server.

        Parameters
        ----------
        path : CloudPath
            The file or folder to remove.
        """
        pass

    @abc.abstractmethod
    def _upload_file(
        self, local_path: Union[str, os.PathLike], cloud_path: BoundedCloudPath
    ) -> BoundedCloudPath:
        pass



def folder_list(folder, recursive):
    """Tests *Client._list_dir function and returns
    the number of items listed
    """
    return {"n_items": len(list(folder.client._list_dir(folder, recursive=recursive)))}



#!/usr/bin/env python

"""The setup script."""

from collections import defaultdict
from setuptools import setup, find_packages
from itertools import chain
from pathlib import Path


def load_requirements(path: Path):
    requirements = defaultdict(list)
    with path.open("r") as fp:
        reqs_type = "base"
        for line in fp.readlines():
            if line.startswith("## extras:"):
                reqs_type = line.partition(":")[-1].strip()
                if reqs_type in ("base", "all"):
                    raise ValueError(f"'{reqs_type}' is a reserved extras keyword.")
            if line.startswith("-r"):
                requirements += load_requirements(line.split(" ")[1].strip())["base"]
            else:
                requirement = line.strip()
                if requirement and not requirement.startswith("#"):
                    requirements[reqs_type].append(requirement)
    return requirements


requirements = load_requirements(Path(__file__).parent / "requirements.txt")
extra_reqs = {k: v for k, v in requirements.items() if k != "base"}
extra_reqs["all"] = list(chain(*extra_reqs.values()))

readme = Path("README.md").read_text(encoding="UTF-8")

setup(
    author="DrivenData",
    author_email="info@drivendata.org",
    python_requires=">=3.7",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description=("pathlib-style classes for cloud storage services"),
    extras_require=extra_reqs,
    install_requires=requirements["base"],
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    name="cloudpathlib",
    packages=find_packages(exclude=["tests"]),
    package_data={"cloudpathlib": ["py.typed"]},
    project_urls={
        "Bug Tracker": "https://github.com/drivendataorg/cloudpathlib/issues",
        "Documentation": "https://cloudpathlib.drivendata.org/",
        "Source Code": "https://github.com/drivendataorg/cloudpathlib",
    },
    url="https://github.com/drivendataorg/cloudpathlib",
    version="0.11.0",
)



import mimetypes
import os
from pathlib import Path, PurePosixPath
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union


from ..client import Client, register_client_class
from ..cloudpath import implementation_registry
from .s3path import S3Path

try:
    from boto3.session import Session
    from boto3.s3.transfer import TransferConfig
    from botocore.config import Config
    from botocore.exceptions import ClientError
    import botocore.session
except ModuleNotFoundError:
    implementation_registry["s3"].dependencies_loaded = False


@register_client_class("s3")
class S3Client(Client):
    """Client class for AWS S3 which handles authentication with AWS for [`S3Path`](../s3path/)
    instances. See documentation for the [`__init__` method][cloudpathlib.s3.s3client.S3Client.__init__]
    for detailed authentication options."""

    def __init__(
        self,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        no_sign_request: Optional[bool] = False,
        botocore_session: Optional["botocore.session.Session"] = None,
        profile_name: Optional[str] = None,
        boto3_session: Optional["Session"] = None,
        local_cache_dir: Optional[Union[str, os.PathLike]] = None,
        endpoint_url: Optional[str] = None,
        boto3_transfer_config: Optional["TransferConfig"] = None,
        content_type_method: Optional[Callable] = mimetypes.guess_type,
    ):
        """Class constructor. Sets up a boto3 [`Session`](
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html).
        Directly supports the same authentication interface, as well as the same environment
        variables supported by boto3. See [boto3 Session documentation](
        https://boto3.amazonaws.com/v1/documentation/api/latest/guide/session.html).

        If no authentication arguments or environment variables are provided, then the client will
        be instantiated as anonymous, which will only have access to public buckets.

        Args:
            aws_access_key_id (Optional[str]): AWS access key ID.
            aws_secret_access_key (Optional[str]): AWS secret access key.
            aws_session_token (Optional[str]): Session key for your AWS account. This is only
                needed when you are using temporarycredentials.
            no_sign_request: (Optional[bool]): If `True`, credentials are not looked for and we use unsigned
                requests to fetch resources. This will only allow access to public resources. This is equivalent
                to `--no-sign-request` in the AWS CLI (https://docs.aws.amazon.com/cli/latest/reference/).
            botocore_session (Optional[botocore.session.Session]): An already instantiated botocore
                Session.
            profile_name (Optional[str]): Profile name of a profile in a shared credentials file.
            boto3_session (Optional[Session]): An already instantiated boto3 Session.
            local_cache_dir (Optional[Union[str, os.PathLike]]): Path to directory to use as cache
                for downloaded files. If None, will use a temporary directory.
            endpoint_url (Optional[str]): S3 server endpoint URL to use for the constructed boto3 S3 resource and client.
                Parameterize it to access a customly deployed S3-compatible object store such as MinIO, Ceph or any other.
            boto3_transfer_config (Optional[dict]): Instantiated TransferConfig for managing s3 transfers.
                (https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig)
            content_type_method (Optional[Callable]): Function to call to guess media type (mimetype) when
                writing a file to the cloud. Defaults to `mimetypes.guess_type`. Must return a tuple (content type, content encoding).
        """
        endpoint_url = endpoint_url or os.getenv("AWS_ENDPOINT_URL")
        if boto3_session is not None:
            self.sess = boto3_session
        else:
            self.sess = Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                botocore_session=botocore_session,
                profile_name=profile_name,
            )

        if no_sign_request:
            self.s3 = self.sess.resource(
                "s3",
                endpoint_url=endpoint_url,
                config=Config(signature_version=botocore.session.UNSIGNED),
            )
            self.client = self.sess.client(
                "s3",
                endpoint_url=endpoint_url,
                config=Config(signature_version=botocore.session.UNSIGNED),
            )
        else:
            self.s3 = self.sess.resource("s3", endpoint_url=endpoint_url)
            self.client = self.sess.client("s3", endpoint_url=endpoint_url)

        self.boto3_transfer_config = boto3_transfer_config

        super().__init__(local_cache_dir=local_cache_dir, content_type_method=content_type_method)

    def _get_metadata(self, cloud_path: S3Path) -> Dict[str, Any]:
        data = self.s3.ObjectSummary(cloud_path.bucket, cloud_path.key).get()

        return {
            "last_modified": data["LastModified"],
            "size": data["ContentLength"],
            "etag": data["ETag"],
            "content_type": data["ContentType"],
            "extra": data["Metadata"],
        }

    def _download_file(self, cloud_path: S3Path, local_path: Union[str, os.PathLike]) -> Path:
        local_path = Path(local_path)
        obj = self.s3.Object(cloud_path.bucket, cloud_path.key)

        obj.download_file(str(local_path), Config=self.boto3_transfer_config)
        return local_path

    def _is_file_or_dir(self, cloud_path: S3Path) -> Optional[str]:
        # short-circuit the root-level bucket
        if not cloud_path.key:
            return "dir"

        # get first item by listing at least one key
        s3_obj = self._s3_file_query(cloud_path)

        if s3_obj is None:
            return None

        # since S3 only returns files when filtering objects:
        # if the first item key is equal to the path key, this is a file
        if s3_obj.key == cloud_path.key:

            # "fake" directories on S3 can be created in the console UI
            # these are 0-size keys that end in `/`
            #  Ref: https://github.com/boto/boto3/issues/377
            if s3_obj.key.endswith("/") and s3_obj.content_length == 0:
                return "dir"
            else:
                return "file"
        else:
            return "dir"

    def _exists(self, cloud_path: S3Path) -> bool:
        # check if this is a bucket
        if not cloud_path.key:
            try:
                self.client.head_bucket(Bucket=cloud_path.bucket)
                return True
            except ClientError:
                return False

        return self._s3_file_query(cloud_path) is not None

    def _s3_file_query(self, cloud_path: S3Path):
        """Boto3 query used for quick checks of existence and if path is file/dir"""
        # check if this is an object that we can access directly
        try:
            obj = self.s3.Object(cloud_path.bucket, cloud_path.key)
            obj.load()
            return obj

        # else, confirm it is a dir by filtering to the first item under the prefix
        except ClientError:
            key = cloud_path.key.rstrip("/") + "/"

            return next(
                (
                    obj
                    for obj in (
                        self.s3.Bucket(cloud_path.bucket).objects.filter(Prefix=key).limit(1)
                    )
                ),
                None,
            )

    def _list_dir(self, cloud_path: S3Path, recursive=False) -> Iterable[Tuple[S3Path, bool]]:
        prefix = cloud_path.key
        if prefix and not prefix.endswith("/"):
            prefix += "/"

        yielded_dirs = set()

        paginator = self.client.get_paginator("list_objects_v2")

        for result in paginator.paginate(
            Bucket=cloud_path.bucket, Prefix=prefix, Delimiter=("" if recursive else "/")
        ):
            # yield everything in common prefixes as directories
            for result_prefix in result.get("CommonPrefixes", []):
                canonical = result_prefix.get("Prefix").rstrip("/")  # keep a canonical form
                if canonical not in yielded_dirs:
                    yield (
                        self.CloudPath(f"s3://{cloud_path.bucket}/{canonical}"),
                        True,
                    )
                    yielded_dirs.add(canonical)

            # check all the keys
            for result_key in result.get("Contents", []):
                # yield all the parents of any key that have not been yielded already
                o_relative_path = result_key.get("Key")[len(prefix) :]
                for parent in PurePosixPath(o_relative_path).parents:
                    parent_canonical = prefix + str(parent).rstrip("/")
                    if parent_canonical not in yielded_dirs and str(parent) != ".":
                        yield (
                            self.CloudPath(f"s3://{cloud_path.bucket}/{parent_canonical}"),
                            True,
                        )
                        yielded_dirs.add(parent_canonical)

                # if we already yielded this dir, go to next item in contents
                canonical = result_key.get("Key").rstrip("/")
                if canonical in yielded_dirs:
                    continue

                # s3 fake directories have 0 size and end with "/"
                if result_key.get("Key").endswith("/") and result_key.get("Size") == 0:
                    yield (
                        self.CloudPath(f"s3://{cloud_path.bucket}/{canonical}"),
                        True,
                    )
                    yielded_dirs.add(canonical)

                # yield object as file
                else:
                    yield (
                        self.CloudPath(f"s3://{cloud_path.bucket}/{result_key.get('Key')}"),
                        False,
                    )

    def _move_file(self, src: S3Path, dst: S3Path, remove_src: bool = True) -> S3Path:
        # just a touch, so "REPLACE" metadata
        if src == dst:
            o = self.s3.Object(src.bucket, src.key)
            o.copy_from(
                CopySource={"Bucket": src.bucket, "Key": src.key},
                Metadata=self._get_metadata(src).get("extra", {}),
                MetadataDirective="REPLACE",
            )

        else:
            target = self.s3.Object(dst.bucket, dst.key)
            target.copy({"Bucket": src.bucket, "Key": src.key})

            if remove_src:
                self._remove(src)
        return dst

    def _remove(self, cloud_path: S3Path, missing_ok: bool = True) -> None:
        try:
            obj = self.s3.Object(cloud_path.bucket, cloud_path.key)

            # will throw if not a file
            obj.load()

            resp = obj.delete()
            assert resp.get("ResponseMetadata").get("HTTPStatusCode") == 204

        except ClientError:
            # try to delete as a direcotry instead
            bucket = self.s3.Bucket(cloud_path.bucket)

            prefix = cloud_path.key
            if prefix and not prefix.endswith("/"):
                prefix += "/"

            resp = bucket.objects.filter(Prefix=prefix).delete()

            # ensure directory deleted; if cloud_path did not exist at all
            # resp will be [], so no need to check success
            if resp:
                assert resp[0].get("ResponseMetadata").get("HTTPStatusCode") == 200
            else:
                if not missing_ok:
                    raise FileNotFoundError(f"File does not exist: {cloud_path}")

    def _upload_file(self, local_path: Union[str, os.PathLike], cloud_path: S3Path) -> S3Path:
        obj = self.s3.Object(cloud_path.bucket, cloud_path.key)

        extra_args = {}

        if self.content_type_method is not None:
            content_type, content_encoding = self.content_type_method(str(local_path))
            if content_type is not None:
                extra_args["ContentType"] = content_type
            if content_encoding is not None:
                extra_args["ContentEncoding"] = content_encoding

        obj.upload_file(str(local_path), Config=self.boto3_transfer_config, ExtraArgs=extra_args)
        return cloud_path


S3Client.S3Path = S3Client.CloudPath  # type: ignore
